/* filtdir.c -- filter file names in directories

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* Akim Demaille <demaille@inf.enst.fr>
 * Based on savedir.c, written by David MacKenzie <djm@gnu.ai.mit.edu>. */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#ifdef CLOSEDIR_VOID
/* Fake a return value. */
#define CLOSEDIR(d) (closedir (d), 0)
#else
#define CLOSEDIR(d) closedir (d)
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifndef errno
extern int errno;
#endif

#if ENABLE_NLS
# include <libintl.h>
# define _(Text) gettext (Text)
#else
# define textdomain(Domain)
# define _(Text) Text
#endif

#include "error.h"
#include "filtdir.h"
#include "xmalloc.h"
#include "xfnmatch.h"

#include <stdio.h>

void
filterdir (const char *dir, struct darray * arr,
	   int (*filter) (const char *))
{
  DIR *dirp;
  struct dirent *dp;

  dirp = opendir (dir);
  if (dirp == NULL)
    return;
/*    error (1, errno, _("cannot open directory `%s'"), dir);*/

  while ((dp = readdir (dirp)) != NULL)
    {
      /* Skip "." and ".." (some NFS filesystems' directories lack them). */
      if ((dp->d_name[0] != '.'
	   || (dp->d_name[1] != '\0'
	       && (dp->d_name[1] != '.' || dp->d_name[2] != '\0')))
	  && (!filter || (*filter) (dp->d_name)))
	da_append (arr, xstrdup (dp->d_name));
    }

  if (CLOSEDIR (dirp))
    error (1, errno, _("cannot close directory `%s'"), dir);
}

/*
 * Same, but the argument is a pattern
 */
void
filterdir_pattern (const char *dir, struct darray * arr,
		   const char * pattern)
{
  DIR *dirp;
  struct dirent *dp;

  dirp = opendir (dir);
  if (dirp == NULL)
    return;
/*    error (1, errno, _("cannot open directory `%s'"), dir);*/

  while ((dp = readdir (dirp)) != NULL)
    {
      /* Skip "." and ".." (some NFS filesystems' directories lack them). */
      if ((dp->d_name[0] != '.'
	   || (dp->d_name[1] != '\0'
	       && (dp->d_name[1] != '.' || dp->d_name[2] != '\0')))
	  && !fnmatch (pattern, dp->d_name, 0))
	da_append (arr, xstrdup (dp->d_name));
    }

  if (CLOSEDIR (dirp))
    error (1, errno, _("cannot close directory `%s'"), dir);
}
